import os
import shutil
import subprocess
import tempfile
from tempfile import mkdtemp
from textwrap import dedent
from typing import Any, Optional, Tuple, Union

from redun.file import File, Staging
from redun.task import Task, task
from redun.utils import iter_nested_value, map_nested_value

NULL = object()
# By default, use bash shell with immediate exit on first error.
DEFAULT_SHELL = "#!/usr/bin/env bash\nset -exo pipefail"


class ScriptError(Exception):
    """
    Error raised when user script returns failure (non-zero exit code).
    """

    def __init__(self, stderr: bytes):
        self.message: Union[bytes, str]

        try:
            self.message = stderr.decode("utf8")
        except UnicodeDecodeError:
            # Error might not be utf8. Keep as is.
            self.message = stderr

    def __str__(self) -> str:
        if isinstance(self.message, str):
            lines = self.message.rstrip("\n").rsplit("\n")
            return "Last line: " + lines[-1]
        else:
            return ""

    def __repr__(self) -> str:
        return f"ScriptError('{str(self)}')"


def prepare_command(command: str, default_shell=DEFAULT_SHELL) -> str:
    """
    Prepare a command string execution by removing surrounding blank lines and dedent.

    Also if an interpreter is not specified, add the default shell as interpreter.
    """
    command = dedent(command).strip()
    if not command.startswith("#!"):
        command = default_shell.rstrip("\n") + "\n" + command
    return command


def get_task_command(task: Task, args: Tuple, kwargs: dict) -> str:
    """
    Get command from a script task.
    """
    command = task.func(*args, **kwargs)
    return prepare_command(command)


def exec_script(command: str) -> bytes:
    """
    Run a script as a subprocess.
    """
    fd, command_file = tempfile.mkstemp()
    try:
        os.write(fd, command.encode("utf8"))
        os.close(fd)

        command2 = """\
chmod +x {command_file}
{command_file}
""".format(
            command_file=command_file
        )
        proc = subprocess.run(
            command2, check=False, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        result, error = proc.stdout, proc.stderr
    finally:
        os.remove(command_file)

    if proc.returncode != 0:
        # Raise error if command had error.
        raise ScriptError(error)

    return result


def get_command_eof(command: str, eof_prefix: str = "EOF") -> str:
    """
    Determine a safe end-of-file keyword to use for a given command to wrap.
    """
    index = 0
    eof = eof_prefix
    lines = command.split("\n")

    while True:
        if eof in lines:
            index += 1
            eof = eof_prefix + str(index)
        else:
            return eof


def get_wrapped_command(command: str, eof_prefix: str = "EOF") -> str:
    """
    Returns a shell script for executing a script written in any language.

    Consider `command` written in python:

    .. code-block:: python

        '''
        #!/usr/bin/env python

        print('Hello, World')
        '''

    In order to turn this into a regular sh shell script, we need to write
    this command to a temporary file, mark the file as executable,
    execute the file, and remove the temporary file.
    """
    wrapped_command = """\
(
# Save command to temp file.
COMMAND_FILE="$(mktemp)"
cat > "$COMMAND_FILE" <<"{eof}"
{command}
{eof}

# Execute temp file.
chmod +x "$COMMAND_FILE"
"$COMMAND_FILE"
RETCODE=$?

# Remove temp file.
rm "$COMMAND_FILE"

exit $RETCODE
)
""".format(
        command=command, eof=get_command_eof(command, eof_prefix=eof_prefix)
    )
    return wrapped_command


@task(name="script_task", namespace="redun", version="1", script=True)
def script_task(command: str) -> str:
    """
    Execute a shell script as redun Task.
    """
    return command


@task(name="script", namespace="redun", version="1", check_valid="shallow")
def _script(
    command: str,
    inputs: Any,
    outputs: Any,
    task_options: dict = {},
    temp_path: Optional[str] = None,
) -> Any:
    """
    Internal task for executing a script.

    This task correctly implements reactivity to changing inputs and outputs.
    `script_task()` alone is unable to implement such reactivity because its
    only argument is a shell script string and its output is the stdout.
    Thus, the ultimate input and output files of the script are accessed
    outside the usual redun detection mechanisms (task arguments
    and return values).

    To achieve the correct reactivity, `script_task()` is special-cased in the Scheduler
    to not use caching, in order to force it to always execute when called.
    Additionally, `_script()` is configured with `check_valid="shallow"` to
    skip execution of its child tasks, `script_task()` and `postprocess_script()`,
    if its previous outputs are still valid (i.e. not altered or deleted).
    """
    # Note: inputs are an argument just for reactivity sake.
    # They have already been incorporated into the command.
    return postprocess_script(
        script_task.options(**task_options)(command), outputs, temp_path=temp_path
    )


@task(name="postprocess_script", namespace="redun", version="1")
def postprocess_script(result: Any, outputs: Any, temp_path: Optional[str] = None) -> Any:
    """
    Postprocess the results of a script task.
    """

    def get_file(value: Any) -> Any:
        if isinstance(value, File) and value.path == "-":
            # File for script stdout.
            return result
        elif isinstance(value, Staging):
            # Staging files and dir turn into their remote versions.
            cls = type(value.remote)
            return cls(value.remote.path)
        else:
            return value

    if temp_path:
        shutil.rmtree(temp_path)

    return map_nested_value(get_file, outputs)


def script(
    command: str, inputs: Any = [], outputs: Any = NULL, tempdir=False, **task_options
) -> Any:
    """
    Execute a shell script as a redun task with file staging.
    """
    if outputs == NULL:
        outputs = File("-")

    command_parts = []

    # Prepare tempdir if requested.
    temp_path: Optional[str]
    if tempdir:
        temp_path = mkdtemp(suffix=".tempdir")
        command_parts.append('cd "{}"'.format(temp_path))
    else:
        temp_path = None

    # Stage inputs.
    command_parts.extend(input.render_stage() for input in iter_nested_value(inputs))

    # User command.
    command_parts.append(get_wrapped_command(prepare_command(command)))

    # Unstage outputs.
    file_stages = [value for value in iter_nested_value(outputs) if isinstance(value, Staging)]
    command_parts.extend(file_stage.render_unstage() for file_stage in file_stages)

    full_command = "\n".join(command_parts)

    # Get input files for reactivity.
    def get_file(value: Any) -> Any:
        if isinstance(value, Staging):
            # Staging files and dir turn into their remote versions.
            cls = type(value.remote)
            return cls(value.remote.path)
        else:
            return value

    input_args = map_nested_value(get_file, inputs)
    return _script(
        full_command, input_args, outputs, task_options=task_options, temp_path=temp_path
    )



import os
import shlex
import shutil

import boto3
import pytest
import vcr

import redun.file
from redun import File, Scheduler, ShardedS3Dataset, task
from redun.file import Dir, StagingDir, StagingFile, get_filesystem, glob_file
from redun.tests.utils import mock_s3, use_tempdir
from redun.tools import copy_file
from redun.value import get_type_registry


@use_tempdir
def test_file_api() -> None:
    file = File("hello.txt")
    file.remove()

    assert not file.exists()

    # Removing a file that doesn't exist should be safe.
    file.remove()

    with file.open("w") as out:
        out.write("hello")

    assert file.exists()

    assert file.read() == "hello"

    hash = file.get_hash()

    with file.open("w") as out:
        out.write("hello2")

    assert file.get_hash() != hash

    file2 = File("hello2.txt")
    assert file.copy_to(file2) == file2
    assert file2.read() == "hello2"

    assert file.size() == 6


@use_tempdir
def test_encoding() -> None:
    """
    File read/write/open should support encodings.
    """
    text = "hello \u1F600"

    # Read and write utf-8.
    file = File("hello.txt")
    file.write(text, encoding="utf-8")
    assert file.read(encoding="utf-8") == text
    assert file.read("rb") == b"hello \xe1\xbd\xa00"

    # Read and write utf-16.
    file2 = File("hello.txt")
    file2.write(text, encoding="utf-16")
    assert file2.read(encoding="utf-16") == text
    assert file2.read("rb") == b"\xff\xfeh\x00e\x00l\x00l\x00o\x00 \x00`\x1f0\x00"

    # Ensure File context can use encoding.
    with file.open(mode="w", encoding="utf-8") as out:
        out.write(text)

    with file.open(mode="r", encoding="utf-8") as infile:
        assert infile.read() == text

    # Binary mode cannot be used with encoding.
    with pytest.raises(ValueError, match="encoding"):
        file.open(mode="rb", encoding="utf-8")


@use_tempdir
def test_dir_api() -> None:

    dir = Dir("hi")
    assert dir.rel_path(os.getcwd()) == ".."

    assert not dir.exists()
    dir.mkdir()
    assert dir.exists()

    f = dir.file("h")
    f.touch()
    assert not dir.is_valid()

    with pytest.raises(OSError):
        dir.rmdir(recursive=False)
    assert dir.exists()

    dir.rmdir(recursive=True)
    assert not dir.exists()
    assert not f.exists()


@use_tempdir
def test_file_subvalues() -> None:
    """
    Files should be discovered as subvalues of a nested value.
    """
    file1 = File("file1")
    file2 = File("file2")
    file3 = File("file3")

    value = [file1, {"a": file2, "b": file3}]
    registry = get_type_registry()
    subvalues = set(registry.iter_subvalues(value))
    assert subvalues == {file1, file2, file3}


@use_tempdir
def test_dir_subvalues() -> None:
    """
    Files should be discovered as subvalues of a Dir.
    """
    file1 = File("dir/file1")
    file1.write("a")
    file2 = File("dir/file2")
    file2.write("a")
    file3 = File("dir/file3")
    file3.write("a")

    dir = Dir("dir")

    # Ensure Dir contains our files.
    file_hashes = {file.hash for file in dir}
    assert file_hashes == {file1.hash, file2.hash, file3.hash}

    registry = get_type_registry()
    subvalues = set(registry.iter_subvalues(dir))
    subvalue_hashes = {subvalue.get_hash() for subvalue in subvalues}
    assert subvalue_hashes == {file1.hash, file2.hash, file3.hash}


@use_tempdir
def test_workflow(scheduler: Scheduler) -> None:

    task_calls = []

    @task()
    def task1(out_path):
        task_calls.append("task1")
        file = File(out_path)
        with file.open("w") as out:
            out.write("hello")
        return file

    @task()
    def task2(infile):
        return "task2({})".format(infile.read())

    @task()
    def workflow():
        file = task1("out.txt")
        return file

    file = scheduler.run(workflow())

    assert task_calls == ["task1"]
    assert file.is_valid()
    assert file.read() == "hello"

    # Rerun a cached workflow.
    file = scheduler.run(workflow())

    assert task_calls == ["task1"]
    assert file.is_valid()
    assert file.read() == "hello"

    # Alter file out-of-band.
    with open(file.path, "w") as out:
        out.write("hello2")
    assert not file.is_valid()

    # Workflow reruns.
    file = scheduler.run(workflow())
    assert file.read() == "hello"
    assert task_calls == ["task1", "task1"]


@use_tempdir
def test_infile(scheduler: Scheduler) -> None:

    task_calls = []

    @task()
    def task1(out_path):
        task_calls.append("task1")
        file = File(out_path)
        with file.open("w") as out:
            out.write("hello")
        return file

    @task()
    def task2(infile):
        task_calls.append("task2")
        return "task2({})".format(infile.read())

    @task()
    def workflow():
        file = task1("out.txt")
        return task2(file)

    data = scheduler.run(workflow())
    assert data == "task2(hello)"
    assert task_calls == ["task1", "task2"]

    data = scheduler.run(workflow())
    assert data == "task2(hello)"
    assert task_calls == ["task1", "task2"]


@use_tempdir
def test_new_file(scheduler: Scheduler) -> None:
    @task()
    def task1(infile):
        return "task1({})".format(infile.read())

    @task()
    def workflow(file):
        return task1(file)

    file = File("out.txt")
    file.write("hello")

    data = scheduler.run(workflow(File("out.txt")))
    assert data == "task1(hello)"

    file.write("hello2")

    data = scheduler.run(workflow(File("out.txt")))
    assert data == "task1(hello2)"


def test_is_valid() -> None:
    """
    File should not be valid if recorded hash does not match current file hash.
    """
    new_file = File("new_file")
    new_file._hash = "123"
    assert not new_file.is_valid()


@use_tempdir
def test_file_hash() -> None:
    file = File("hello.txt")
    file.write("hello")
    assert file.is_valid()


def test_glob() -> None:
    expected = [
        "redun/tests/test_data/file_glob/",
        "redun/tests/test_data/file_glob/aa",
        "redun/tests/test_data/file_glob/aa/bb",
        "redun/tests/test_data/file_glob/aa/bb/cc.txt",
        "redun/tests/test_data/file_glob/aa/bb/dd.txt",
        "redun/tests/test_data/file_glob/aa/zz.txt",
    ]
    assert sorted(glob_file("redun/tests/test_data/file_glob/**")) == expected


@use_tempdir
def test_dir() -> None:
    src_dir = os.path.join(os.path.dirname(__file__), "test_data")

    path = "base"
    shutil.copytree(src_dir, path)

    # Directory should be valid initially.
    dir = Dir(path)
    assert dir.is_valid()

    # Adding a new file should cause the Dir to not be valid anymore.
    File(path + "/qqq.txt").touch()
    assert not dir.is_valid()


@use_tempdir
def test_check_valid_result(scheduler: Scheduler) -> None:
    file = File("output")
    file.write("hello")

    @task()
    def task1():
        return File("output")

    @task()
    def task2():
        return task1().read()

    @task()
    def main(check_valid="full"):
        return task2.options(check_valid=check_valid)()

    assert scheduler.run(main()) == "hello"

    file.write("hello2")

    # The result of task1, which is deeper down the call stack, will have
    # its result checked, and be re-evaluated.
    assert scheduler.run(main()) == "hello2"

    # Using valid_check=shallow, we don't re-evaluate task1.
    file.write("hello3")
    assert scheduler.run(main("shallow")) == "hello2"


@use_tempdir
def test_check_valid_args() -> None:
    file = File("input")
    file.write("hello")

    @task()
    def task1(file):
        return file.read()

    @task()
    def task2():
        file = File("input")
        return task1(file)

    @task()
    def main(check_valid="full"):
        return task2.options(check_valid=check_valid)()

    scheduler = Scheduler()
    assert scheduler.run(main()) == "hello"

    file.write("hello2")

    # The result of task1, which is deeper down the call stack, will have
    # its result checked, and be re-evaluated.
    assert scheduler.run(main()) == "hello2"

    # Create a new scheduler, with new cache.
    scheduler = Scheduler()
    assert scheduler.run(main("shallow")) == "hello2"

    # Using valid_check=shallow, we don't re-evaluate task1.
    file.write("hello3")
    assert scheduler.run(main("shallow")) == "hello2"


def test_staging_file() -> None:
    """
    StagingFile constructor should support paths and Files.
    """

    # Path-based constructor.
    staging_file = StagingFile("local.txt", "remote.txt")
    assert isinstance(staging_file.local, File)
    assert isinstance(staging_file.remote, File)
    assert staging_file.local.path == "local.txt"
    assert staging_file.remote.path == "remote.txt"

    # File-based constructor.
    staging_file2 = StagingFile(File("local.txt"), File("remote.txt"))
    assert isinstance(staging_file2.local, File)
    assert isinstance(staging_file2.remote, File)
    assert staging_file2.local.path == "local.txt"
    assert staging_file2.remote.path == "remote.txt"


def test_file_stage() -> None:
    """
    File.stage() should produce a StagingFile.
    """

    file = File("remote/file.txt")
    staging_file = file.stage("local/file.txt")
    assert staging_file.local.path == "local/file.txt"
    assert staging_file.remote.path == "remote/file.txt"

    staging_file = file.stage("local/")
    assert staging_file.local.path == "local/file.txt"

    staging_file = file.stage()
    assert staging_file.local.path == "file.txt"


@use_tempdir
def test_staging_self() -> None:
    """
    StagingFile/StagingDir should allow staging to itself.
    """
    file = File("file.txt")
    file.write("hello")

    staging_file = file.stage("file.txt")
    assert staging_file.render_stage() == ""
    assert staging_file.render_unstage() == ""

    file2 = staging_file.stage()
    assert file2.hash == file.hash

    file2 = staging_file.unstage()
    assert file2.hash == file.hash

    dir = Dir("dir")
    dir.file("a").write("a")
    dir.file("b").write("b")
    dir.mkdir()

    staging_dir = dir.stage("dir")
    assert staging_dir.render_stage() == ""
    assert staging_dir.render_unstage() == ""

    dir2 = staging_dir.stage()
    assert dir2.hash == dir.hash

    dir2 = staging_dir.unstage()
    assert dir2.hash == dir.hash


def test_staging_dir() -> None:
    """
    StagingDir constructor should support paths and Files.
    """

    # Path-based constructor.
    staging_dir = StagingDir("local", "remote")
    assert isinstance(staging_dir.local, Dir)
    assert isinstance(staging_dir.remote, Dir)
    assert staging_dir.local.path == "local"
    assert staging_dir.remote.path == "remote"

    # File-based constructor.
    staging_dir2 = StagingDir(Dir("local"), Dir("remote"))
    assert isinstance(staging_dir2.local, Dir)
    assert isinstance(staging_dir2.remote, Dir)
    assert staging_dir2.local.path == "local"
    assert staging_dir2.remote.path == "remote"


def test_dir_stage() -> None:
    """
    Dir.stage() should produce a StagingDir.
    """

    dir = Dir("remote/dir")
    staging_dir = dir.stage("local/dir")
    assert staging_dir.local.path == "local/dir"
    assert staging_dir.remote.path == "remote/dir"

    staging_dir = dir.stage()
    assert staging_dir.local.path == "dir"


@use_tempdir
def test_staging_file_value() -> None:
    """
    StagingFile should implement the Value interface.
    """
    remote = File("remote.txt")
    staging_file = remote.stage("local.txt")
    assert isinstance(staging_file, StagingFile)
    assert StagingFile.type_name == "redun.StagingFile"

    # StagingFile should satisfy the Value interface.
    assert staging_file.get_hash() == "cfb6d4d9de4d45ed80bbdb6d411ffc22618b9db1"

    data = staging_file.serialize()
    staging_file2 = StagingFile.deserialize(StagingFile, data)

    assert staging_file2.remote.path == "remote.txt"
    assert staging_file2.local.path == "local.txt"

    assert staging_file.is_valid()

    # The hash of a StagingFile should only depend on the paths of the File pair,
    # not on the contents (hash) of the Files.
    remote.write("hello")
    assert staging_file.get_hash() == "cfb6d4d9de4d45ed80bbdb6d411ffc22618b9db1"


@mock_s3
@use_tempdir
def test_sharded_dataset_from_file() -> None:
    # Create a S3 bucket with a few files.
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    # Raise an error if files don't exist
    with pytest.raises(FileNotFoundError):
        _ = ShardedS3Dataset.from_files(
            [
                File("s3://example-bucket/bar/baz.csv"),
                File("s3://example-bucket/x/y.csv"),
                File("s3://example-bucket/bar/baz/baz2.csv"),
            ]
        )

    file1 = File("s3://example-bucket/data/more_data/shard1.csv")
    file2 = File("s3://example-bucket/data/other_data/shard2.csv")
    file3 = File("s3://example-bucket/data/shard3.csv")
    file4 = File("s3://example-bucket/data/some.parquet")
    file1.write("word1, word2\nhello, world")
    file2.write("word1, word2\ncarrots, potatoes")
    file3.write("word1, word2\nfoo, bar")
    file4.write("well it looks like a parquet file")

    # Test a single file
    x = ShardedS3Dataset.from_files([file1], allow_additional_files=True)
    assert x.path == "s3://example-bucket/data/more_data"
    assert x.format == "csv"

    # Test incompatible file formats.
    with pytest.raises(ValueError):
        _ = ShardedS3Dataset.from_files([file1, file2, file3, file4])

    # Test files not on S3
    local_file = File("./potato.csv")
    local_file.write("i like potato")
    with pytest.raises(ValueError):
        _ = ShardedS3Dataset.from_files([local_file])

    # Test empty file list
    with pytest.raises(ValueError):
        _ = ShardedS3Dataset.from_files([])

    file4.remove()

    # Making a dataset that would contain additional files (in this case file3)
    # should raise an error.
    with pytest.raises(ValueError):
        _ = ShardedS3Dataset.from_files([file1, file2])

    x = ShardedS3Dataset.from_files([file1, file2], allow_additional_files=True)
    assert len(x.filenames) == 3
    assert file3.path in x.filenames
    assert x.format == "csv"
    assert x.path == "s3://example-bucket/data"

    x = ShardedS3Dataset.from_files([file1, file2, file3])
    assert len(x.filenames) == 3
    assert file3.path in x.filenames
    assert x.format == "csv"
    assert x.path == "s3://example-bucket/data"


@mock_s3
def test_sharded_dataset() -> None:
    """
    Tests ShardedS3Dataset hashing
    """
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    # We have 2 shards, 1 partition with another shard, and a txt file.
    file1 = File("s3://example-bucket/data/shard1.csv")
    file1.write("word1, word2\nhello, world")
    file2 = File("s3://example-bucket/data/shard2.csv")
    file2.write("word1, word2\ncarrots, potatoes")
    file3 = File("s3://example-bucket/data/word1=foo/shard1.csv")
    file3.write("word1, word2\nfoo, bar")
    decoy = File("s3://example-bucket/data/not_a_csv.txt")
    decoy.write("i am not valid input")

    dataset = ShardedS3Dataset(path="s3://example-bucket/data", format="csv", recurse=True)
    dataset2 = ShardedS3Dataset(path="s3://example-bucket/data", format="csv", recurse=True)

    # Check the correct files are pulled in.
    assert sorted(dataset.filenames) == sorted([file1.path, file2.path, file3.path])
    assert sorted(dataset2.filenames) == sorted([file1.path, file2.path, file3.path])

    # Update the recurse value and ensure the list of files is updated.
    dataset.recurse = False
    assert sorted(dataset.filenames) == sorted([file1.path, file2.path])

    # Check hashing changes when list of files changes.
    assert dataset.hash == "3394197d206ea0ef46795131b98f86c52ab9a508"
    assert dataset2.hash == "f7cd6b0188ff900f0ca00ea0b937a91d70a3e67a"

    file3.remove()
    assert dataset.is_valid()
    assert not dataset2.is_valid()

    # Changing the path should update the list of files.
    dataset2.path = "s3://example-bucket/nonexistent"
    assert dataset2.filenames == []

    # Changing the format should as well.
    dataset2.path = "s3://example-bucket/data"
    assert len(dataset2.filenames)
    dataset2.format = "parquet"
    assert dataset2.filenames == []

    # Load should not work without glue context.
    with pytest.raises(ValueError):
        _ = dataset.load_spark()

    # Try loading one shard.
    dat = dataset.load_pandas_shards(max_shards=1)
    assert len(dat) == 1
    assert isinstance(dat, list)

    # Now load the whole dataset.
    concatdat = dataset.load_pandas()
    assert concatdat[concatdat["word1"] == "carrots"].shape == (1, 2)
    assert list(concatdat["word1"].sort_values()) == ["carrots", "hello"]


@use_tempdir
def test_staging_dir_value() -> None:
    """
    StagingDir should implement the Value interface.
    """
    remote = Dir("remote")
    staging_dir = remote.stage("local")
    assert isinstance(staging_dir, StagingDir)
    assert StagingDir.type_name == "redun.StagingDir"

    # StagingDir should satisfy the Value interface.
    assert staging_dir.get_hash() == "bb933a6873e68cdb47ecf5af8c5784608966ae14"

    data = staging_dir.serialize()
    staging_dir2 = StagingDir.deserialize(StagingDir, data)

    assert staging_dir2.remote.path == "remote"
    assert staging_dir2.local.path == "local"

    assert staging_dir.is_valid()

    # The hash of a StagingDir should only depend on the paths of the File pair,
    # not on the contents (hash) of the Files.
    File("remote/file").write("hello")
    assert staging_dir.get_hash() == "bb933a6873e68cdb47ecf5af8c5784608966ae14"


@use_tempdir
@mock_s3
def test_staging_render() -> None:
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    file = File("s3://example-bucket/remote.txt")
    file.write("hello")

    # Ensure StagingFile is setup correctly.
    staging_file = file.stage("local.txt")
    assert isinstance(staging_file, StagingFile)
    assert staging_file.remote.path == "s3://example-bucket/remote.txt"
    assert staging_file.local.path == "local.txt"

    # Stage a file.
    assert not staging_file.local.exists()
    staging_file.stage()
    assert staging_file.local.read() == "hello"

    # Unstage a file.
    staging_file.local.write("hello2")
    staging_file.unstage()
    assert staging_file.remote.read() == "hello2"

    # S3 staging commands.
    assert (
        staging_file.render_stage()
        == "aws s3 cp --no-progress s3://example-bucket/remote.txt local.txt"
    )
    assert (
        staging_file.render_unstage()
        == "aws s3 cp --no-progress local.txt s3://example-bucket/remote.txt"
    )

    # Local filesystem staging commands.
    file2 = File("remote.txt")
    staging_file2 = file2.stage("local.txt")
    assert staging_file2.render_stage() == "cp remote.txt local.txt"
    assert staging_file2.render_unstage() == "cp local.txt remote.txt"


def test_staging_render_gs() -> None:
    """
    Should be able to render stage commands for gs.
    """
    cmd = File("gs://bucket/remote.txt").stage("local.txt").render_stage()
    assert cmd == "gsutil cp gs://bucket/remote.txt local.txt"

    cmd = File("gs://bucket/remote.txt").stage("local.txt").render_unstage()
    assert cmd == "gsutil cp local.txt gs://bucket/remote.txt"


def test_render_stage_quote() -> None:
    file = File("s3://example-bucket/crazy name \" \\ ' .txt")
    stage_file = file.stage()
    command = stage_file.render_stage()
    assert command == (
        r"""aws s3 cp --no-progress 's3://example-bucket/crazy name " \ '"'"' .txt' 'crazy name " \ '"'"' .txt'"""  # noqa: E501
    )
    assert shlex.split(command) == [
        "aws",
        "s3",
        "cp",
        "--no-progress",
        stage_file.remote.path,
        stage_file.local.path,
    ]


@use_tempdir
@mock_s3
def test_staging_dir_render() -> None:
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    File("s3://example-bucket/dir/a.txt").write("a")
    File("s3://example-bucket/dir/b.txt").write("b")
    File("s3://example-bucket/dir/c/d.txt").write("d")

    # Ensure StagingDir is setup correctly.
    dir = Dir("s3://example-bucket/dir")
    staging_dir = dir.stage("local")

    assert isinstance(staging_dir, StagingDir)
    assert staging_dir.remote.path == "s3://example-bucket/dir"
    assert staging_dir.local.path == "local"

    # Stage a dir.
    assert not staging_dir.local.exists()
    staging_dir.stage()
    set(staging_dir.local) == {
        "s3://example-bucket/dir/a.txt",
        "s3://example-bucket/dir/b.txt",
        "s3://example-bucket/dir/c/d.txt",
    }

    # Unstage a dir.
    staging_dir.local.file("dir/a.txt").write("hello2")
    staging_dir.unstage()
    staging_dir.remote.file("dir/a.txt").read() == "hello2"

    # S3 staging commands.
    assert (
        staging_dir.render_stage()
        == "aws s3 cp --no-progress --recursive s3://example-bucket/dir local"
    )
    assert (
        staging_dir.render_unstage()
        == "aws s3 cp --no-progress --recursive local s3://example-bucket/dir"
    )

    # Local filesystem staging commands.
    dir2 = Dir("remote")
    staging_dir2 = dir2.stage("local")
    assert staging_dir2.render_stage() == "cp -r remote local"
    assert staging_dir2.render_unstage() == "cp -r local remote"


@use_tempdir
def test_copy_file_task(scheduler: Scheduler) -> None:
    """
    copy_file should copy a file from one location to another.
    """
    src_file = File("data.txt")
    src_file.write("hello")

    dest_file = scheduler.run(copy_file(src_file, "data2.txt"))
    assert dest_file.read() == "hello"


@vcr.use_cassette("redun/tests/test_data/vcr_cassettes/test_https.yaml")
@use_tempdir
def test_https() -> None:
    """
    File should work with https URLs.
    """
    url = "https://www.google.com/robots.txt"
    expected_data = "User-agent"

    # Use FileSystem for https protocol.
    fs = get_filesystem(proto="https")
    assert fs.open(url, "r").read(10) == expected_data

    # Open and read a File over https.
    file = File(url)
    with file.open() as infile:
        assert infile.read(10) == expected_data

    # https Files should support hashing.
    assert file.get_hash() == "106920f9bb975ad607f701992c6ca855f9309c96"

    # Should support file size.
    assert file.size() > 10

    # We should be able to copy from https to local files.
    local_file = File("robots.txt")
    file.copy_to(local_file)
    assert local_file.open().read(10) == expected_data

    # Other File methods should work as expected.
    assert fs.isfile(url)
    assert not fs.isdir(url)


def test_shell_copy_local() -> None:
    """
    LocalFileSystem should be able to generate shell commands for file copies.
    """
    assert File("src.txt").shell_copy_to("dest.txt") == "cp src.txt dest.txt"
    assert File("src.txt").shell_copy_to("dest space.txt") == "cp src.txt 'dest space.txt'"
    assert File("src.txt").shell_copy_to(None) == "cat src.txt"
    assert get_filesystem("local").shell_copy(None, "dest.txt") == "cat > dest.txt"

    assert get_filesystem("local").shell_copy("src", "dest", recursive=True) == "cp -r src dest"


def test_shell_copy_s3() -> None:
    """
    S3FileSystem should be able to generate shell commands for file copies.
    """
    assert (
        File("s3://bucket/src.txt").shell_copy_to("s3://bucket/dest.txt")
        == "aws s3 cp --no-progress s3://bucket/src.txt s3://bucket/dest.txt"
    )
    assert (
        File("s3://bucket/src.txt").shell_copy_to("dest.txt")
        == "aws s3 cp --no-progress s3://bucket/src.txt dest.txt"
    )
    assert (
        File("src.txt").shell_copy_to("s3://bucket/dest.txt")
        == "aws s3 cp --no-progress src.txt s3://bucket/dest.txt"
    )
    assert (
        File("s3://bucket/src.txt").shell_copy_to(None)
        == "aws s3 cp --no-progress s3://bucket/src.txt -"
    )
    assert (
        get_filesystem("s3").shell_copy(None, "s3://bucket/dest.txt")
        == "aws s3 cp --no-progress - s3://bucket/dest.txt"
    )

    assert (
        get_filesystem("s3").shell_copy("s3://bucket/src", "s3://bucket/dest", recursive=True)
        == "aws s3 cp --no-progress --recursive s3://bucket/src s3://bucket/dest"
    )
    with pytest.raises(ValueError):
        assert get_filesystem("s3").shell_copy("s3://bucket/src", None, recursive=True)


def test_shell_copy_gs() -> None:
    """
    GSFileSystem should be able to generate shell commands for file copies.
    """

    assert (
        File("gs://bucket/src.txt").shell_copy_to("gs://bucket/dest.txt")
        == "gsutil cp gs://bucket/src.txt gs://bucket/dest.txt"
    )
    assert (
        File("gs://bucket/src.txt").shell_copy_to("dest.txt")
        == "gsutil cp gs://bucket/src.txt dest.txt"
    )
    assert (
        File("src.txt").shell_copy_to("gs://bucket/dest.txt")
        == "gsutil cp src.txt gs://bucket/dest.txt"
    )
    assert File("gs://bucket/src.txt").shell_copy_to(None) == "gsutil cp gs://bucket/src.txt -"
    assert (
        get_filesystem("gs").shell_copy(None, "gs://bucket/dest.txt")
        == "gsutil cp - gs://bucket/dest.txt"
    )

    assert (
        get_filesystem("gs").shell_copy("gs://bucket/src", "gs://bucket/dest", recursive=True)
        == "gsutil cp -r gs://bucket/src gs://bucket/dest"
    )
    with pytest.raises(ValueError):
        assert get_filesystem("gs").shell_copy("gs://bucket/src", None, recursive=True)


def test_shell_copy_fs() -> None:
    """
    FileSystem should be able to generate shell commands for file copies.
    """
    # Cross cloud shell copies.
    assert (
        File("s3://bucket/src.txt").shell_copy_to("gs://bucket/dest.txt")
        == "(aws s3 cp --no-progress s3://bucket/src.txt -) | (gsutil cp - gs://bucket/dest.txt)"
    )
    assert (
        File("http://example.com/src.txt").shell_copy_to("gs://bucket/dest.txt")
        == "redun fs cp http://example.com/src.txt gs://bucket/dest.txt"
    )

    assert (
        Dir("s3://bucket/src").shell_copy_to("gs://bucket/dest")
        == "redun fs cp --recursive s3://bucket/src gs://bucket/dest"
    )


@use_tempdir
def test_copy_file() -> None:
    """
    Copy a file with copy_file().
    """
    src_file = File("src.txt")
    src_file.write("hello")
    dest_file = File("dest.txt")

    redun.file.copy_file(src_file.path, dest_file.path)
    assert dest_file.read() == src_file.read()


@use_tempdir
def test_copy_file_dir() -> None:
    """
    Copy a directory with copy_file().
    """
    a = File("src/a.txt")
    a.write("a")
    b = File("src/b/b.txt")
    b.write("b")

    dest_dir = Dir("dest")

    redun.file.copy_file("src", dest_dir.path, recursive=True)
    assert {file.path for file in dest_dir} == {"dest/a.txt", "dest/b/b.txt"}



import abc
import datetime
from base64 import b64decode, b64encode
from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional

import sqlalchemy as sa
from dateutil.parser import parse as parse_date
from sqlalchemy.orm import joinedload, selectinload
from sqlalchemy.orm.query import Query

from redun.backends import db
from redun.backends.base import TagEntityType

# Serialization version.
VERSION = 1


def serialize_timestamp(timestamp: Optional[datetime.datetime]) -> Optional[str]:
    """
    Serialize a datetime value to a timestamp string.
    """
    if timestamp:
        return str(timestamp)
    else:
        return None


def deserialize_timestamp(date_str: Optional[str]) -> Optional[datetime.datetime]:
    """
    Deserialize timestamp string to datetime value.
    """
    if date_str:
        return parse_date(date_str)
    else:
        return None


class Serializer(abc.ABC):
    """
    Base class serializer for RedunBackendDb records.
    """

    pk_field: str = ""

    def __init__(self, db_version: "Optional[db.DBVersionInfo]" = None):
        self._db_version = db_version or db.REDUN_DB_VERSIONS[-1]

    def get_pk(self, spec: dict) -> str:
        return spec[self.pk_field]

    def serialize(self, row: "db.Base") -> dict:
        raise NotImplementedError()

    def serialize_query(self, query: Query) -> Iterator[dict]:
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        raise NotImplementedError()


class ExecutionSerializer(Serializer):
    pk_field = "id"

    def serialize(self, execution: "db.Base") -> dict:
        return {
            "_version": VERSION,
            "_type": "Execution",
            "id": execution.id,
            "args": execution.args,
            "job_id": execution.job_id,
        }

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Execution"
        return [
            db.Execution(
                id=spec["id"],
                args=spec["args"],
                job_id=spec["job_id"],
            )
        ]


class JobSerializer(Serializer):
    pk_field = "id"

    def serialize(self, job: "db.Base") -> dict:
        return {
            "_version": VERSION,
            "_type": "Job",
            "id": job.id,
            "start_time": serialize_timestamp(job.start_time),
            "end_time": serialize_timestamp(job.end_time),
            "task_hash": job.task_hash,
            "cached": job.cached,
            "status": job.status,
            "call_hash": job.call_hash,
            "parent_id": job.parent_id,
            "children": [child.id for child in job.child_jobs],
            "execution_id": job.execution_id,
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:

        # Fetch parent-child links.
        ChildJob = sa.orm.aliased(db.Job)
        parent_child_pairs = (
            query.from_self(db.Job.id, ChildJob.id)
            .join(ChildJob, ChildJob.parent_id == db.Job.id)
            .order_by(ChildJob.start_time)
            .all()
        )
        parent2children = defaultdict(list)
        for parent_id, child_id in parent_child_pairs:
            parent2children[parent_id].append(child_id)

        for job in query.all():
            yield {
                "_version": VERSION,
                "_type": "Job",
                "id": job.id,
                "start_time": serialize_timestamp(job.start_time),
                "end_time": serialize_timestamp(job.end_time),
                "task_hash": job.task_hash,
                "cached": job.cached,
                "status": job.status,
                "call_hash": job.call_hash,
                "parent_id": job.parent_id,
                "children": parent2children[job.id],
                "execution_id": job.execution_id,
            }

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Job"
        return [
            db.Job(
                id=spec["id"],
                start_time=deserialize_timestamp(spec["start_time"]),
                end_time=deserialize_timestamp(spec["end_time"]),
                task_hash=spec["task_hash"],
                cached=spec["cached"],
                call_hash=spec["call_hash"],
                parent_id=spec["parent_id"],
                execution_id=spec["execution_id"],
            )
        ]


class ValueSerializer(Serializer):
    pk_field = "value_hash"

    def serialize(self, value: "db.Base") -> dict:
        # Up to one special subtype should exist at a time
        assert not (bool(value.file) and bool(value.task))
        subtype: Optional[Dict[str, Any]] = None
        if value.file:
            subtype = {
                "_type": "File",
                "path": value.file.path,
            }
        elif value.task:
            subtype = {
                "_type": "Task",
                "name": value.task.name,
                "namespace": value.task.namespace,
                "source": value.task.source,
            }
        return {
            "_version": VERSION,
            "_type": "Value",
            "value_hash": value.value_hash,
            "type": value.type,
            "format": value.format,
            "data": b64encode(value.value).decode("utf8"),
            "subvalues": sorted(edge.value_hash for edge in value.child_edges),
            "subtype": subtype,
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:
        query = query.options(
            joinedload(db.Value.file),
            joinedload(db.Value.task),
            joinedload(db.Value.child_edges),
        )
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Value"
        return (
            [
                db.Value(
                    value_hash=spec["value_hash"],
                    type=spec["type"],
                    format=spec["format"],
                    value=b64decode(spec["data"]),
                ),
            ]
            + [
                db.Subvalue(parent_value_hash=spec["value_hash"], value_hash=child_id)
                for child_id in spec["subvalues"]
            ]
            + (
                [
                    db.File(
                        value_hash=spec["value_hash"],
                        path=spec["subtype"]["path"],
                    )
                ]
                if spec.get("subtype") and spec["subtype"]["_type"] == "File"
                else []
            )
            + (
                [
                    db.Task(
                        hash=spec["value_hash"],
                        name=spec["subtype"]["name"],
                        namespace=spec["subtype"]["namespace"],
                        source=spec["subtype"]["source"],
                    )
                ]
                if spec.get("subtype") and spec["subtype"]["_type"] == "Task"
                else []
            )
        )


class CallNodeSerializer(Serializer):
    pk_field = "call_hash"

    def serialize(self, call_node: "db.Base") -> dict:
        args = {
            arg.arg_key
            or str(arg.arg_position): {
                "arg_hash": arg.arg_hash,
                "value_hash": arg.value_hash,
                "upstream": sorted(arg_result.result_call_hash for arg_result in arg.arg_results),
            }
            for arg in call_node.arguments
        }

        return {
            "_version": VERSION,
            "_type": "CallNode",
            "call_hash": call_node.call_hash,
            "task_name": call_node.task_name,
            "task_hash": call_node.task_hash,
            "args_hash": call_node.args_hash,
            "value_hash": call_node.value_hash,
            "timestamp": serialize_timestamp(call_node.timestamp),
            "args": args,
            "children": [edge.child_id for edge in call_node.child_edges],
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:
        query = query.options(
            selectinload(db.CallNode.arguments).joinedload(db.Argument.arg_results),
            selectinload(db.CallNode.child_edges),
        )
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "CallNode"

        return (
            [
                db.CallNode(
                    call_hash=spec["call_hash"],
                    task_name=spec["task_name"],
                    task_hash=spec["task_hash"],
                    args_hash=spec["args_hash"],
                    value_hash=spec["value_hash"],
                    timestamp=deserialize_timestamp(spec["timestamp"]),
                )
            ]
            + [
                db.CallEdge(
                    parent_id=spec["call_hash"],
                    child_id=child_hash,
                    call_order=call_index,
                )
                for call_index, child_hash in enumerate(spec["children"])
            ]
            + [
                db.Argument(
                    arg_hash=arg["arg_hash"],
                    call_hash=spec["call_hash"],
                    value_hash=arg["value_hash"],
                    arg_position=int(key) if key.isdigit() else None,
                    arg_key=key if not key.isdigit() else None,
                )
                for key, arg in spec["args"].items()
            ]
            + [
                db.ArgumentResult(
                    arg_hash=arg["arg_hash"],
                    result_call_hash=upstream,
                )
                for key, arg in spec["args"].items()
                for upstream in arg["upstream"]
            ]
        )


class TagSerializer(Serializer):
    pk_field = "tag_hash"

    def serialize(self, tag: "db.Base") -> dict:
        return {
            "_version": VERSION,
            "_type": "Tag",
            "tag_hash": tag.tag_hash,
            "entity_type": tag.entity_type.name,
            "entity_id": tag.entity_id,
            "key": tag.key,
            "value": tag.value,
            "parents": [edit.parent_id for edit in tag.parent_edits],
        }

    def serialize_query(self, query: Query) -> Iterator[dict]:
        query = query.options(
            selectinload(db.Tag.parent_edits).joinedload(db.TagEdit.child),
        )
        for row in query.all():
            yield self.serialize(row)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        assert spec["_version"] == VERSION
        assert spec["_type"] == "Tag"

        return [
            db.Tag(
                tag_hash=spec["tag_hash"],
                entity_type=TagEntityType(spec["entity_type"]),
                entity_id=spec["entity_id"],
                key=spec["key"],
                value=spec["value"],
            )
        ] + [db.TagEdit(parent_id=parent, child_id=spec["tag_hash"]) for parent in spec["parents"]]


class RecordSerializer(Serializer):
    def __init__(self, db_version: "Optional[db.DBVersionInfo]" = None):
        super().__init__(db_version)
        self._serializers = {
            "Execution": ExecutionSerializer(db_version),
            "Job": JobSerializer(db_version),
            "CallNode": CallNodeSerializer(db_version),
            "Value": ValueSerializer(db_version),
            "Tag": TagSerializer(db_version),
        }

    def get_pk(self, spec: dict) -> str:
        return self._serializers[spec["_type"]].get_pk(spec)

    def get_subserializer(self, type: str) -> Serializer:
        return self._serializers[type]

    def serialize(self, obj: "db.Base") -> dict:
        if isinstance(obj, db.Task):
            # Use Value-version of Task for serialization.
            obj = obj.value
        return self._serializers[type(obj).__name__].serialize(obj)

    def serialize_query(self, query: Query) -> Iterator[dict]:
        [column] = query.column_descriptions
        return self._serializers[column["name"]].serialize_query(query)

    def deserialize(self, spec: dict) -> "List[db.Base]":
        return self._serializers[spec["_type"]].deserialize(spec)



import os
from configparser import SectionProxy

import pytest

from redun.cli import get_config_dir, postprocess_config, setup_scheduler
from redun.config import Config
from redun.tests.utils import use_tempdir


def test_config_parse_sections():
    config_string = """
[executors.batch]
type = aws_batch
image = 123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3
queue = queue
s3_scratch = s3://example-bucket/redun/

[executors.batch2]
type = aws_batch
image = 123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3
queue = queue2
s3_scratch = s3://example-bucket/redun/
"""

    config = Config()
    config.read_string(config_string)

    assert sorted(config.keys()) == ["executors"]
    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["batch"]["queue"] == "queue"
    assert config["executors"]["batch2"]["type"] == "aws_batch"
    assert config["executors"]["batch2"]["queue"] == "queue2"


@pytest.fixture
def sample_config():
    return Config(
        {
            "executors.batch": {
                "type": "aws_batch",
                "image": "123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3",
                "queue": "queue",
                "s3_scratch": "s3://example-bucket/redun/",
            },
            "executors.batch2": {
                "type": "aws_batch",
                "image": "123.abc.ecr.us-west-2.amazonaws.com/amazonlinux-python3",
                "queue": "queue2",
                "s3_scratch": "s3://example-bucket/redun/",
            },
        }
    )


def test_config_dict(sample_config):
    config = sample_config

    assert sorted(config.keys()) == ["executors"]
    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["batch"]["queue"] == "queue"
    assert config["executors"]["batch2"]["type"] == "aws_batch"
    assert config["executors"]["batch2"]["queue"] == "queue2"


def test_get_config_dict(sample_config):
    """Assert get_config_dict() reproduces a Config correctly"""
    config = Config(config_dict=sample_config.get_config_dict())
    assert sorted(config.keys()) == ["executors"]
    assert config["executors"]["batch"]["type"] == "aws_batch"
    assert config["executors"]["batch"]["queue"] == "queue"
    assert config["executors"]["batch2"]["type"] == "aws_batch"
    assert config["executors"]["batch2"]["queue"] == "queue2"


@use_tempdir
def test_get_config_dict_replace_config_dir(sample_config):
    """Verify replace_config_dir behavior"""
    os.makedirs(".redun")
    with open(".redun/redun.ini", "w") as out:
        out.write(
            """
[backend]
db_uri = sqlite:///redun.db
"""
        )

    scheduler = setup_scheduler()

    # Ensure initial config dir and files are created.
    assert os.path.exists(".redun/redun.ini")
    assert os.path.exists(".redun/redun.db")

    conf = scheduler.config.get_config_dict()
    local_config_dir = get_config_dir()
    db_uri = conf["backend"]["db_uri"]
    assert db_uri == f"sqlite:///{local_config_dir}/redun.db"
    assert conf["backend"]["config_dir"] == local_config_dir
    assert conf["repos.default"]["config_dir"] == local_config_dir

    replacement = "/tmp/xyz"
    conf = scheduler.config.get_config_dict(replace_config_dir=replacement)
    db_uri = conf["backend"]["db_uri"]
    assert db_uri == f"sqlite:///{replacement}/redun.db"
    assert conf["backend"]["config_dir"] == replacement
    assert conf["repos.default"]["config_dir"] == replacement


def test_postprocess_config(sample_config):
    """
    Default sections must be true Sections.
    If they aren't an exception will occur when get_config_dict() is called
    """
    post_processed_config = postprocess_config(sample_config, "/tmp/config_dir")
    config = Config(config_dict=post_processed_config.get_config_dict())

    assert sorted(config.keys()) == ["backend", "executors", "repos"]
    config = postprocess_config(config, "/tmp/config_dir")
    # Default repository should be added since it was missing
    assert config["repos"]["default"]["config_dir"] == "/tmp/config_dir"
    # db_uri should be converted to absolute absolute config_dir
    assert config["backend"]["db_uri"] == "sqlite:////tmp/config_dir/redun.db"

    # Ensure that the auto-added sections are truly Section objects and not plain dicts.
    assert isinstance(config["repos"]["default"], SectionProxy)
    assert isinstance(config["backend"], SectionProxy)



import os
from typing import Dict, List, Optional, TypeVar

from redun.file import Dir, File
from redun.functools import seq
from redun.task import task

T = TypeVar("T")


@task(namespace="redun", version="1", cache=False)
def debug_task(value: T) -> T:
    """
    Launch pdb debugger during one point of the redun workflow.

    For example, to inspect the lazy value `x` in the workflow below
    we can pass it through `debug_task()`:

    .. code-block:: python

        x = task1()
        x = debug_task(x)
        y = task2(x)
    """
    import pdb

    pdb.set_trace()
    return value


@task(namespace="redun", version="1", config_args=["skip_if_exists"])
def copy_file(src_file: File, dest_path: str, skip_if_exists: bool = False) -> File:
    """
    Copy a File to a new path.

    This task can help in parallelizing file copies and providing data
    provenance.
    """
    if dest_path.endswith("/"):
        # If dest_path is a directory, use the same basename as the src_file.
        dest_path = dest_path + src_file.basename()

    return src_file.copy_to(File(dest_path), skip_if_exists=skip_if_exists)


@task(namespace="redun", version="1", config_args=["skip_if_exists", "copy_options"])
def copy_files(
    src_files: List[File],
    dest_paths: List[str],
    skip_if_exists: bool = False,
    copy_options: Dict = {},
) -> List[File]:
    """
    Copy multiple Files to new paths.

    File copies will be done in parallel.
    """
    return [
        copy_file.options(**copy_options)(src_file, dest_path, skip_if_exists=skip_if_exists)
        for src_file, dest_path in zip(src_files, dest_paths)
    ]


@task(namespace="redun", version="1", config_args=["skip_if_exists", "copy_options"])
def copy_dir(
    src_dir: Dir, dest_path: str, skip_if_exists: bool = False, copy_options: Dict = {}
) -> Dir:
    """
    Copy a Dir to a new path.

    File copies will be done in parallel.
    """
    src_files = list(src_dir)
    dest_dir = Dir(dest_path)
    dest_paths = [dest_dir.file(src_dir.rel_path(src_file.path)).path for src_file in src_files]
    return seq(
        [
            copy_files(
                src_files, dest_paths, skip_if_exists=skip_if_exists, copy_options=copy_options
            ),
            dest_dir,
        ]
    )[1]


@task(namespace="redun", version="1")
def render_template(
    out_path: str, template: File, context: dict, template_dir: Optional[str] = None
) -> File:
    """
    Render a jinja2 template.
    """
    try:
        import jinja2
    except ImportError:
        raise ImportError("jinja2 must be installed to use render_template()")

    if not template_dir:
        # Infer template directory from template File.
        template_dir = template.dirname()

    env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_dir))
    jinja_template = env.get_template(os.path.relpath(template.path, template_dir))

    # Render template.
    out_file = File(out_path)
    out_file.write(jinja_template.render(context))
    return out_file



import json
import pickle
from base64 import b64decode, b64encode
from collections import defaultdict
from datetime import datetime
from typing import Callable, Dict, Iterable, List, Tuple, cast
from unittest.mock import Mock, patch

from sqlalchemy.orm import Session

from redun import File as RedunFile
from redun import Scheduler, apply_tags, task
from redun.backends.base import TagEntityType
from redun.backends.db import (
    Base,
    CallEdge,
    CallNode,
    Execution,
    Job,
    RedunBackendDb,
    Tag,
    TagEdit,
    Value,
    parse_db_version,
)
from redun.backends.db.serializers import RecordSerializer
from redun.tests.utils import listen_queries
from redun.utils import pickle_dumps

DB_VERSIONS = [
    parse_db_version("3.0"),
    parse_db_version("3.1"),
]


def serialize(obj: Base) -> dict:
    return RecordSerializer().serialize(obj)


def deserialize(spec: dict) -> List[Base]:
    return RecordSerializer().deserialize(spec)


def dump_record(record: dict) -> str:
    """
    Helper method for comparing records.
    """
    if "children" in record:
        record["children"].sort()
    if "parents" in record:
        record["parents"].sort()
    return json.dumps(record, sort_keys=True)


def group_edges(edges: Iterable[Tuple[str, Base, str]]) -> Dict[str, List[str]]:
    """
    Group serialization edges by edge type.
    """
    edge_groups = defaultdict(list)
    for edge_type, _, record_id in edges:
        edge_groups[edge_type].append(record_id)
    return edge_groups


def test_serializers():
    # Value.
    value = Value(
        value_hash="222abc",
        type="builtins.int",
        format="application/python-pickle",
        value=pickle_dumps(10),
    )
    expected_spec = {
        "_version": 1,
        "_type": "Value",
        "value_hash": "222abc",
        "type": "builtins.int",
        "format": "application/python-pickle",
        "data": b64encode(pickle_dumps(10)).decode("utf8"),
        "subvalues": [],
        "subtype": None,
    }
    assert serialize(value) == expected_spec
    assert serialize(deserialize(expected_spec)[0]) == expected_spec

    # CallNode.
    call_node = CallNode(
        call_hash="333abc",
        task_name="ns.task1",
        task_hash="123abc",
        args_hash="444abc",
        value_hash="222abc",
        timestamp=datetime(2020, 1, 1),
    )
    child_node = CallNode(
        call_hash="555abc",
        task_name="ns.task1",
        task_hash="123abc",
        args_hash="444abc",
        value_hash="222abc",
        timestamp=datetime(2020, 1, 1),
    )
    expected_spec = {
        "_version": 1,
        "_type": "CallNode",
        "call_hash": "333abc",
        "task_name": "ns.task1",
        "task_hash": "123abc",
        "args_hash": "444abc",
        "value_hash": "222abc",
        "timestamp": "2020-01-01 00:00:00",
        "args": {},
        "children": ["555abc"],
    }

    with patch(
        "redun.backends.db.CallNode.child_edges",
        new_callable=lambda: [Mock(child_id=child_node.call_hash)],
    ):
        spec = serialize(call_node)
    assert spec == expected_spec

    [call_node2, edge] = deserialize(expected_spec)
    assert isinstance(call_node2, CallNode)
    assert isinstance(edge, CallEdge)

    with patch(
        "redun.backends.db.CallNode.child_edges",
        new_callable=lambda: [Mock(child_id=child_node.call_hash)],
    ):
        spec = serialize(call_node2)
    assert spec == expected_spec


def test_special_value_serialization(scheduler: Scheduler) -> None:
    @task(version="2")
    def main(file: RedunFile):
        return file

    f = RedunFile("/tmp/file1.txt")

    scheduler.run(main(f))

    [file_spec, task_spec] = list(scheduler.backend.get_records([f.hash, main.hash], sorted=True))
    expected_file_spec = {
        "_version": 1,
        "_type": "Value",
        "value_hash": f.hash,
        "type": "redun.File",
        "format": "application/python-pickle",
        "data": b64encode(pickle_dumps(f)).decode("utf8"),
        "subvalues": [],
        "subtype": {"_type": "File", "path": "/tmp/file1.txt"},
    }
    expected_task_spec = {
        "_version": 1,
        "_type": "Value",
        "value_hash": main.hash,
        "type": "redun.Task",
        "format": "application/python-pickle",
        "data": b64encode(pickle_dumps(main)).decode("utf8"),
        "subvalues": [],
        "subtype": {
            "_type": "Task",
            "name": "main",
            "namespace": "",
            "source": "    def main(file: RedunFile):\n        return file\n",
        },
    }
    assert file_spec == expected_file_spec
    assert task_spec == expected_task_spec


def test_serialize_tags() -> None:
    """
    Ensure Tags can be serialized and deserialized.
    """
    tag = Tag(
        tag_hash="123",
        entity_type=TagEntityType.Value,
        entity_id="1",
        key="env",
        value=["prod", "stg"],
    )
    parent_tag = Tag(
        tag_hash="222",
        entity_type=TagEntityType.Value,
        entity_id="1",
        key="env",
        value=["prod"],
    )

    with patch(
        "redun.backends.db.Tag.parent_edits",
        new_callable=lambda: [Mock(parent_id=parent_tag.tag_hash)],
    ):
        spec = serialize(tag)
    assert spec == {
        "_type": "Tag",
        "_version": 1,
        "entity_id": "1",
        "entity_type": "Value",
        "key": "env",
        "tag_hash": "123",
        "value": ["prod", "stg"],
        "parents": ["222"],
    }

    [tag2, tag_edit] = deserialize(spec)
    assert isinstance(tag2, Tag)
    assert isinstance(tag_edit, TagEdit)

    with patch(
        "redun.backends.db.Tag.parent_edits",
        new_callable=lambda: [Mock(parent_id=parent_tag.tag_hash)],
    ):
        spec2 = serialize(tag2)
    assert spec == spec2


def test_get_records(scheduler: Scheduler) -> None:
    @task()
    def task1(x):
        return x + 1

    @task()
    def main(tsk: Callable[[int], int]):
        return [tsk(1), tsk(2)]

    # Record some tasks and calls.
    scheduler.run(main(task1))

    # Fetch records.
    assert isinstance(scheduler.backend, RedunBackendDb)
    [record] = list(scheduler.backend.get_records([main.hash]))
    assert record["value_hash"] == main.hash

    # Transfer record to another backend.
    scheduler2 = Scheduler()
    assert isinstance(scheduler2.backend, RedunBackendDb)
    assert list(scheduler2.backend.get_records([main.hash])) == []
    scheduler2.backend.put_records([record])
    [record2] = list(scheduler2.backend.get_records([main.hash]))
    assert record2 == record


def test_iter_record_ids(scheduler: Scheduler, backend: RedunBackendDb, session: Session) -> None:
    @task()
    def task1():
        return 1

    @task()
    def task2(x):
        return [x, task1()]

    @task()
    def main():
        return [task2(task1()), task2(2)]

    # Record some tasks and calls.
    scheduler.run(main())

    # Get execution.
    execution = session.query(Execution).one()
    [execution_record] = list(backend.get_records([execution.id]))
    assert execution_record["_type"] == "Execution"

    # Get the job edge.
    [(edge_name, model, job_id)] = [
        (edge_name, model, job_id)
        for edge_name, model, job_id in backend.get_child_record_ids([(Execution, execution.id)])
        if model == Job
    ]

    assert edge_name == "Execution.job"
    assert job_id == execution_record["job_id"]
    [job] = list(backend.get_records([job_id]))
    assert job["_type"] == "Job"

    [
        (_, _, task_hash),
        (_, _, call_hash),
        (_, _, job_id1),
        (_, _, job_id2),
        (_, _, job_id3),
    ] = list(backend.get_child_record_ids([(Job, job_id)]))

    assert task_hash == main.hash

    # Get root call node.
    call_node = session.query(CallNode).filter_by(task_name="main").first()
    assert call_hash == call_node.call_hash

    # Fetch records.
    edge_groups = group_edges(backend.get_child_record_ids([(CallNode, call_node.call_hash)]))

    [value_hash] = edge_groups["CallNode.result"]
    [value_record] = backend.get_records([value_hash])

    assert pickle.loads(b64decode(value_record["data"])) == [[1, 1], [2, 1]]

    # Get task2 children.
    child_id1, child_i2, child_id3 = sorted(edge_groups["CallNode.child_call_node"])
    edge_groups2 = group_edges(backend.get_child_record_ids([(CallNode, child_id1)]))
    edge_groups3 = group_edges(backend.get_child_record_ids([(CallNode, child_id3)]))

    # Upstream call_node should be a child record as well.
    assert (
        list(backend.get_records([edge_groups2["CallNode.child_call_node"][0]]))[0]["task_name"]
        == "task1"
    )

    # Grandchildren should be task1.
    assert (
        edge_groups2["CallNode.child_call_node"][0] == edge_groups3["CallNode.child_call_node"][0]
    )

    record_ids = list(backend.iter_record_ids([execution.id]))
    records = list(backend.get_records(record_ids))

    scheduler2 = Scheduler()
    backend2 = cast(RedunBackendDb, scheduler2.backend)
    backend2.put_records(reversed(records))

    record_ids2 = backend2.iter_record_ids([execution.id])
    records2 = list(backend2.get_records(record_ids2))

    def dumps(record):
        if "children" in record:
            record["children"].sort()
        if "parents" in record:
            record["parents"].sort()
        return json.dumps(record, sort_keys=True)

    # Deep equal.
    assert set(map(dumps, records)) == set(map(dumps, records2))

    # Puts are idempotent.
    backend2.put_records(records)

    record_ids2 = backend2.iter_record_ids([execution.id])
    records2 = list(backend2.get_records(record_ids2))

    # Deep equal.
    assert set(map(dumps, records)) == set(map(dumps, records2))


def test_get_records_queries(
    scheduler: Scheduler, backend: RedunBackendDb, session: Session
) -> None:
    """
    iter_record_ids and get_records should use O(1) queries.
    """

    @task()
    def task1(i):
        return i + 1

    @task()
    def task2(x):
        return [x, task1(x)]

    @task()
    def main(n):
        return [task2(task1(i)) for i in range(n)]

    # Record some tasks and calls.
    scheduler.run(main(1))

    # Get latest execution.
    execution1 = (
        session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(Job.start_time.desc())
        .first()
    )

    with listen_queries(backend.engine) as id_queries1:
        record_ids = list(backend.iter_record_ids([execution1.id]))

    with listen_queries(backend.engine) as record_queries1:
        records1 = list(backend.get_records(record_ids))

    # Record 10x more tasks and calls.
    scheduler.run(main(10))

    # Get latest execution.
    execution2 = (
        session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(Job.start_time.desc())
        .first()
    )

    with listen_queries(backend.engine) as id_queries2:
        record_ids = list(backend.iter_record_ids([execution2.id]))

    with listen_queries(backend.engine) as record_queries2:
        records2 = list(backend.get_records(record_ids))

    # Even though we have ~10x more data in the CallGraph, we should use O(1)
    # queries to fetch it.
    assert len(id_queries1) == len(id_queries2)
    assert len(record_queries1) == len(record_queries2)
    assert len(records2) > len(records1)


def test_sync_tags(scheduler: Scheduler, backend: RedunBackendDb, session: Session) -> None:
    @task(tags=[("env", "prod")])
    def task1():
        return apply_tags(10, [("user", "alice"), ("release", "final")])

    assert scheduler.run(task1(), tags=[("project", "acme")]) == 10
    execution = session.query(Execution).one()

    # Update a tag.
    backend.update_tags(TagEntityType.Execution, execution.id, ["project"], [("project", "skunk")])

    # Delete a tag.
    value = session.query(Value).filter(Value.type == "builtins.int").one()
    backend.delete_tags(value.value_hash, [("release", "final")])

    assert backend.get_tags([execution.id])[execution.id]["project"] == ["skunk"]

    record_ids = list(backend.iter_record_ids([execution.id]))
    records = list(backend.get_records(record_ids))
    tags = [record for record in records if record["_type"] == "Tag"]

    assert any(
        tag["entity_type"] == "Execution"
        and tag["key"] == "project"
        and tag["value"] == "skunk"
        and len(tag["parents"]) == 1
        for tag in tags
    )
    assert any(
        tag["entity_type"] == "Execution"
        and tag["key"] == "project"
        and tag["value"] == "acme"
        and len(tag["parents"]) == 0
        for tag in tags
    )
    assert any(tag["entity_type"] == "Value" and tag["key"] == "user" for tag in tags)
    assert any(tag["entity_type"] == "Task" and tag["key"] == "env" for tag in tags)
    assert any(tag["entity_type"] == "Job" and tag["key"] == "env" for tag in tags)

    # Sync records to second backend.
    scheduler2 = Scheduler()
    backend2 = cast(RedunBackendDb, scheduler2.backend)
    backend2.put_records(reversed(records))

    # Export records from second backend.
    record_ids2 = backend2.iter_record_ids([execution.id])
    records2 = list(backend2.get_records(record_ids2))
    tags2 = [record for record in records2 if record["_type"] == "Tag"]

    # Ensure all tags are present.
    assert set(map(dump_record, tags)) == set(map(dump_record, tags2))

    # Tag update should also be synced.
    assert backend2.get_tags([execution.id])[execution.id]["project"] == ["skunk"]

    # Tag delete should also be synced.
    assert backend2.get_tags([value.value_hash])[value.value_hash] == {"user": ["alice"]}



import logging
import os
import sys
import time
from typing import Iterator, Optional, Tuple, cast

import pytest
from _pytest.monkeypatch import MonkeyPatch
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session

from redun import Scheduler
from redun.backends.db import RedunBackendDb
from redun.config import Config
from redun.task import get_task_registry
from redun.utils import clear_import_paths

logger = logging.getLogger(__name__)


def connect_with_retries(
    uri: str, interval: float = 0.1, timeout: float = 10.0
) -> Tuple[Engine, Connection]:
    """
    Connect to database with retries.
    """
    logger.info("Connecting to test database ({})...".format(uri))
    credentialed_uri = RedunBackendDb._get_credentialed_uri(uri, {})
    engine = create_engine(credentialed_uri)
    stop = time.time() + timeout
    while time.time() < stop:
        try:
            conn = engine.connect()
            logger.info("Connected to test database ({})".format(uri))
            return engine, conn
        except OperationalError:
            time.sleep(interval)

    raise RuntimeError("Unable to connect to test database: {}".format(uri))


@pytest.fixture(scope="session")
def monkeysession(request):
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="session")
def testdb(monkeysession) -> Iterator[Optional[str]]:
    """
    Create a test database.

    By using scope=session the test database is created only once per testing
    session.
    """
    host = "localhost"
    port = 5432
    database_name = "redun"

    monkeysession.setenv("REDUN_DB_USERNAME", "postgres")
    monkeysession.setenv("REDUN_DB_PASSWORD", "postgres")

    test_db_uri_stub = "postgresql://{host}:{port}/".format(
        host=host,
        port=port,
    )
    test_db_uri = test_db_uri_stub + database_name

    if not os.environ.get("REDUN_TEST_POSTGRES"):
        # Postgres database disabled.
        yield None
        return

    # Wait for database to accept connections.
    engine, conn = connect_with_retries(test_db_uri_stub)

    # Create test database.
    conn.execute("commit")
    conn.execute("drop database if exists {}".format(database_name))
    conn.execute("commit")
    conn.execute("create database {}".format(database_name))
    conn.close()
    engine.dispose()

    yield test_db_uri


@pytest.fixture
def scheduler(testdb: Optional[str]) -> Iterator[Scheduler]:
    """
    Returns a Scheduler for testing.
    """
    config: Optional[Config]

    if testdb:
        config = Config(
            {
                "scheduler": {"ignore_warnings": "namespace"},
                "backend": {"db_uri": testdb, "automigrate": "True"},
            }
        )
        scheduler = Scheduler(config=config)
        scheduler.load()
        backend = cast(RedunBackendDb, scheduler.backend)
        assert backend.engine and backend.session
        session = backend.session

        # Run tests inside a transaction, so that we rollback effects.
        # start the session in a SAVEPOINT...
        trans = session.begin(subtransactions=True)
        session.begin_nested()

        # then each time that SAVEPOINT ends, reopen it
        # https://gist.github.com/zzzeek/8443477#file-gistfile1-py-L38
        @event.listens_for(session, "after_transaction_end")
        def restart_savepoint(session, transaction):
            if transaction.nested and not transaction._parent.nested:
                session.begin_nested()

        yield scheduler

        trans.rollback()

        # Explicitly dispose the db engine so we can force kill the docker
        # process without error messages.
        backend.session.close()
        backend.engine.dispose()
        del scheduler

    else:
        yield Scheduler(
            config=Config(
                {
                    "scheduler": {"ignore_warnings": "namespace"},
                }
            )
        )


@pytest.fixture
def backend(scheduler: Scheduler) -> RedunBackendDb:
    """
    Returns a RedunBackend cast as RedunBackendDb.
    """
    backend = cast(RedunBackendDb, scheduler.backend)
    return backend


@pytest.fixture
def session(backend: RedunBackendDb) -> Session:
    """
    Returns a sqlalchemy Session for the redun backend.
    """
    assert backend.session
    return backend.session


@pytest.fixture(autouse=True)
def redun_globals(request):
    """
    pytest fixture for resetting redun global state during tests.

    - python modules
    - python sys.path
    - redun tasks
    - redun import paths
    """
    init_module_names = set(sys.modules.keys())
    init_sys_path = list(sys.path)
    task_registry = get_task_registry()

    # Temporarily remove unrelated tests from the task registry to prevent interference
    # (Note: Resetting the task registry to its original state at the end of this function doesn't
    # work because pytest loads all code modules before running any tests which means at this
    # point, all tasks across all unit tests are present in the task registry.)
    tasks_to_hide = {}
    for task_fullname in list(task_registry._tasks):
        if task_fullname.startswith("redun."):
            continue
        if task_fullname.startswith(request.module.__name__):
            continue
        if hasattr(request.module, "redun_namespace") and task_fullname.startswith(
            request.module.redun_namespace
        ):
            continue
        tasks_to_hide[task_fullname] = task_registry.get(task_fullname)
        del task_registry._tasks[task_fullname]

    yield

    # Reset imported modules.
    for module_name in list(sys.modules.keys()):
        if module_name not in init_module_names:
            sys.modules.pop(module_name)

    # Reset sys.path.
    sys.path = init_sys_path

    # Reset redun extra import paths.
    clear_import_paths()

    # Restore temporarily deleted task
    for k, v in tasks_to_hide.items():
        task_registry._tasks[k] = v



import io
import os
from typing import List
from unittest import mock

import boto3
import pytest

from redun import File
from redun.cli import (
    DEFAULT_REDUN_INI,
    REDUN_CONFIG_ENV,
    REDUN_INI_FILE,
    RedunClient,
    RedunClientError,
)
from redun.tests.utils import assert_match_lines, mock_s3, use_tempdir


def make_config(config_path: str) -> None:
    config_file = File(os.path.join(config_path, REDUN_INI_FILE))
    with config_file.open("w") as out:
        out.write(DEFAULT_REDUN_INI.format(db_uri="sqlite:///other.db"))


def run_command(client: RedunClient, argv: List[str]) -> str:
    """
    Run redun cli command and return output as a string.
    """
    client.stdout = io.StringIO()
    client.execute(argv)
    client.stdout.seek(0)
    return client.stdout.read()


def test_cli_repos() -> None:
    """
    redun push/pull should detect nonexistent repos.
    """
    client = RedunClient()
    client.stdout = io.StringIO()

    # Look for a repo that doesn't exist
    with pytest.raises(RedunClientError):
        client.execute(["redun", "push", "--repo", "nonexistent"])

    with pytest.raises(RedunClientError):
        client.execute(["redun", "pull", "--repo", "nonexistent"])


@use_tempdir
@mock.patch.dict(os.environ, {REDUN_CONFIG_ENV: "."})
@mock_s3
def test_add_list_s3_repos() -> None:
    """
    Tests adding and listing repos from S3.
    """
    client = RedunClient()

    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-repo")

    file = File("s3://example-repo/potato/redun.ini")
    file.write(DEFAULT_REDUN_INI.format(db_uri="sqlite:///other.db"))

    with pytest.raises(FileNotFoundError):
        client.execute(["redun", "repo", "add", "carrots", "s3://example-repo/carrots"])

    client.execute(["redun", "repo", "add", "potato", "s3://example-repo/potato"])

    output = run_command(client, ["redun", "repo", "list"]).split("\n")
    assert "  potato: s3://example-repo/potato" in output


@use_tempdir
@mock.patch.dict(os.environ, {REDUN_CONFIG_ENV: "."})
def test_add_list_remove_repo() -> None:
    """
    Tests adding and removing repos from the config file.
    """
    client = RedunClient()
    os.mkdir("the_repo")

    # No redun.ini in directory
    with pytest.raises(FileNotFoundError):
        client.execute(["redun", "repo", "add", "foo", "the_repo"])
    repos = run_command(client, ["redun", "repo", "list"])
    assert "foo" not in repos

    # Create the config, then it should work
    make_config("the_repo")
    client.execute(["redun", "repo", "add", "foo", "the_repo"])
    repos = run_command(client, ["redun", "repo", "list"])
    assert "the_repo" in repos

    # Duplicate additions should not work.
    with pytest.raises(RedunClientError):
        client.execute(["redun", "repo", "add", "foo", "the_repo"])

    # Adding repository under a different name should print a message.
    output = run_command(client, ["redun", "repo", "add", "bar", "the_repo"])
    assert output == f"Existing repository 'foo' also uses config at {os.getcwd()}/the_repo\n"

    # Adding default repository under a different name should print a message.
    output = run_command(client, ["redun", "repo", "add", "oops", "."])
    assert output == f"Existing repository 'default' also uses config at {os.getcwd()}\n"

    # Repositories cannot be named 'default'.
    with pytest.raises(RedunClientError):
        client.execute(["redun", "repo", "add", "default", "the_repo"])

    # Removing nonexistent repo should not work.
    with pytest.raises(RedunClientError):
        client.execute(["redun", "repo", "remove", "nonexistent"])

    # Removing repo isn't fully implemented.
    with pytest.raises(RedunClientError):
        client.execute(["redun", "repo", "remove", "foo"])


@use_tempdir
@mock.patch.dict(os.environ, {REDUN_CONFIG_ENV: "."})
def test_repo_sync() -> None:
    """
    redun push/pull should synchronize the expected number of records.
    """
    file = File("workflow.py")
    file.write(
        """
from redun import task
redun_namespace = 'test'

@task()
def task1(x: int):
    return x + 1

@task()
def carrots(x: int):
    return f"I have {x} carrots"
    """
    )

    # Set up another repo
    os.mkdir("the_repo")
    make_config("the_repo")
    other_client = RedunClient()
    other_client.execute(["redun", "-c", "the_repo", "init"])

    client = RedunClient()
    client.execute(["redun", "repo", "add", "other", "the_repo"])

    # Run a job locally and in other repo.
    client.execute(["redun", "--repo", "other", "run", "workflow.py", "carrots", "--x", "10"])
    client.execute(["redun", "run", "workflow.py", "task1", "--x", "12"])

    # Commands should not work on nonexistent repos.
    with pytest.raises(RedunClientError):
        client.execute(["redun", "--repo", "nonexistent", "log"])
    with pytest.raises(RedunClientError):
        client.execute(["redun", "push", "nonexistent"])
    with pytest.raises(RedunClientError):
        client.execute(["redun", "pull", "nonexistent"])

    # Check logs.
    lines = run_command(client, ["redun", "log"]).split("\n")
    assert_match_lines(
        [
            r"Recent executions:",
            r"",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py task1 --x 12 \(.*\)",
            r"",
        ],
        lines,
    )
    lines_other = run_command(client, ["redun", "--repo", "other", "log"]).split("\n")
    assert_match_lines(
        [
            r"Recent executions:",
            r"",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  "
            r"--repo other run workflow.py carrots --x 10 \(.*\)",
            r"",
        ],
        lines_other,
    )

    # Sync repos local -> other
    output = run_command(client, ["redun", "pull", "other"])
    assert output == "Pulled 9 new record(s) from repo 'other'\n"

    # Should now have cached value from sync.
    client.execute(["redun", "run", "workflow.py", "carrots", "--x", "10"])
    lines = run_command(client, ["redun", "log", "--task"]).split("\n")
    assert_match_lines(
        ["Task ........ test.task1 ", "Task ........ test.carrots ", ""],
        lines,
    )

    # Should not be able to sync a repo to itself.
    with pytest.raises(RedunClientError):
        client.execute(["redun", "--repo", "other", "push", "other"])

    # Now push to the remote
    output = run_command(client, ["redun", "push", "other"])
    assert output == "Pushed 14 record(s) to repo 'other'\n"
    lines = run_command(client, ["redun", "--repo", "other", "log", "--exec"]).split("\n")
    assert_match_lines(
        [
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  "
            r"run workflow.py carrots --x 10 \(.*\)",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  run workflow.py task1 --x 12 \(.*\)",
            r"Exec ........ \[ DONE \] ....-..-.. ..:..:..:  "
            r"--repo other run workflow.py carrots --x 10 \(.*\)",
            "",
        ],
        lines,
    )
    # Subsequent push should sync 0 new records.
    output = run_command(client, ["redun", "push", "other"])
    assert output == "Repo 'other' is up to date.\n"



"""Add indexes

Revision ID: 71ec303c90e4
Revises: 30ffbaee18cd
Create Date: 2021-07-18 13:21:51.711159

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "71ec303c90e4"
down_revision = "30ffbaee18cd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        "ix_call_node_call_hash_vpo",
        "call_node",
        ["call_hash"],
        unique=True,
        postgresql_ops={"call_hash": "varchar_pattern_ops"},
    )
    op.create_index(
        "ix_execution_id_vpo",
        "execution",
        ["id"],
        unique=True,
        postgresql_ops={"id": "varchar_pattern_ops"},
    )
    op.create_index(op.f("ix_file_path"), "file", ["path"], unique=False)
    op.create_index(
        "ix_job_id_vpo", "job", ["id"], unique=True, postgresql_ops={"id": "varchar_pattern_ops"}
    )
    op.create_index(
        "ix_task_hash_vpo",
        "task",
        ["hash"],
        unique=True,
        postgresql_ops={"hash": "varchar_pattern_ops"},
    )
    op.create_index(
        "ix_task_name_vpo",
        "task",
        ["name"],
        unique=False,
        postgresql_ops={"name": "varchar_pattern_ops"},
    )
    op.create_index(
        "ix_task_namespace_vpo",
        "task",
        ["namespace"],
        unique=False,
        postgresql_ops={"namespace": "varchar_pattern_ops"},
    )
    op.create_index(
        "ix_value_value_hash_vpo",
        "value",
        ["value_hash"],
        unique=True,
        postgresql_ops={"value_hash": "varchar_pattern_ops"},
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "ix_value_value_hash_vpo",
        table_name="value",
        postgresql_ops={"value_hash": "varchar_pattern_ops"},
    )
    op.drop_index(
        "ix_task_namespace_vpo",
        table_name="task",
        postgresql_ops={"namespace": "varchar_pattern_ops"},
    )
    op.drop_index(
        "ix_task_name_vpo", table_name="task", postgresql_ops={"name": "varchar_pattern_ops"}
    )
    op.drop_index(
        "ix_task_hash_vpo", table_name="task", postgresql_ops={"hash": "varchar_pattern_ops"}
    )
    op.drop_index("ix_job_id_vpo", table_name="job", postgresql_ops={"id": "varchar_pattern_ops"})
    op.drop_index(op.f("ix_file_path"), table_name="file")
    op.drop_index(
        "ix_execution_id_vpo", table_name="execution", postgresql_ops={"id": "varchar_pattern_ops"}
    )
    op.drop_index(
        "ix_call_node_call_hash_vpo",
        table_name="call_node",
        postgresql_ops={"call_hash": "varchar_pattern_ops"},
    )
    # ### end Alembic commands ###



from mypy import api

from redun.tests.utils import get_test_file


def test_task_types() -> None:
    """
    mypy should find type errors related to redun task calls.
    """
    workflow_file = get_test_file("test_data/typing/workflow_fail.py.txt")

    stdout, stderr, ret_code = api.run(
        [
            "--show-traceback",
            workflow_file,
            "redun",
        ]
    )
    print(stdout)
    assert ret_code == 1

    # Parse found type check errors.
    stdout_lines = stdout.split("\n")[:-2]
    found_errors = {line.split(":", 2)[1] for line in stdout_lines}

    # Get lines with expected errors.
    with open(workflow_file) as infile:
        expected_errors = {str(i) for i, line in enumerate(infile, 1) if "ERROR" in line}

    assert found_errors == expected_errors
